/* This is part of dview, a dvi previewer. (c) 1995 Thomas Moor          */
/*                                                                       */
/* This program may be used without any warranty. It may be modified and */
/* distributed without any restrictions.                                 */

#include <stdio.h>
#include <math.h>
#include <vga.h>
#include <vgagl.h>

#include "defsx.h"                        

/* from the file globals.c we need ... */
extern FILE* prot;
extern char* dviname;
extern short cpagenum;
extern bigbitmap pxlmem;
extern int vgaxdim, vgaydim;
extern float lrandmm, rrandmm, orandmm, urandmm;
extern int lrandpxl, rrandpxl, orandpxl, urandpxl;
extern int frameon;
extern BMUNIT* bit_masks;
extern uchar* bit_count;
extern uchar* greytab;
extern uchar shrink;  /* stets <= 10 ? */
extern uchar colors;
extern int dvixpos, dviypos;

/* from the file subs.c w need ... */
extern void print_bigbitmap();
extern Tptr(void*);
extern void* findmem(long);

/* from the asemmbler hack memcpy.s we need */
extern mymemcpyflip( long vgaoffset, void *src, int n );
extern mymemcpyflip4( long vgaoffset, void *src, int n );
extern mymemcpyflip1( long vgaoffset, void *src, int n, int skip );
extern mymemcpystd( long vgaoffset, void *src, int n );
extern mymemcpystd4( long vgaoffset, void *src, int n );
extern mymemcpystd1( long vgaoffset, void *src, int n, int skip );


/* globals used by my painting stuff */
static int advixpos, adviypos;  
static GraphicsContext *physicalscreen;
static void *font;
static int thegraphmode, actgraphmode, oldgraphmode;



#define BORDERCOL         COLORS_PER_GREY
#define TEXTCOL           COLORS_PER_GREY+1
#define TEXTBACKCOL       COLORS_PER_GREY+2
#define FRAMECOL	  COLORS_PER_GREY+3

#define FONTSIZE 8                   /* choose 8 or 16 */
#define STATUSLINE 10

/* #define FLIPPIXELS */            /* those are set from the commandline */
/* #define USEMEMCPY  */


/* Function prototypes */
void setpalette (void);
void preparefont(void);
void my_write( int, int, char*); 


/* this became display dependent. the pxlmem should be accesable by the dvi-
   encoding part and by the routines that come with the display. */

void free_bigbitmap(void) {
  short i;  

  /*fprintf(prot,"(freebig"); */
  if(pxlmem.bits != NULL) {
    for(i=0;i<pxlmem.h;i++)
      if(pxlmem.bits[i] != NULL) free(pxlmem.bits[i]);     
    free(pxlmem.bits);
    pxlmem.bits=NULL;
  }
  /*fprintf(prot,")");*/
}


void alloc_bigbitmap(void) {
  short size, i;
  
  /* fprintf(prot,"(allocbig"); */
  pxlmem.bits=(BMUNIT**) findmem(pxlmem.h*sizeof(BMUNIT*));
  memset(pxlmem.bits,0,pxlmem.h*sizeof(BMUNIT*));
  pxlmem.bmu_wide = ROUNDUP(pxlmem.w*pxlmem.type, BITS_PER_BMUNIT);
  size = MAX(pxlmem.bmu_wide*BYTES_PER_BMUNIT,1);
  for(i=0;i<pxlmem.h;i++) 
    pxlmem.bits[i]= (BMUNIT*)findmem(size);
  setpalette();
 /* fprintf(prot,")"); */
}


void vgaclose(void) {
    vga_setmode(oldgraphmode);
    if(font != gl_font8x8) free(font);
}

void vgaerror(char* mssg)
{
   fprintf(prot,"\nfatal error: display: %s\n",mssg);   
   vgaclose();
   exit(1);
}

void vgaopen(void) {
  int n;
  uchar g;
  
  vga_init();
  actgraphmode=vga_getcurrentmode();
  oldgraphmode=actgraphmode;
  n=0;
  if(vgaxdim==320 && vgaydim==240) n=5;
  if(vgaxdim==640 && vgaydim==480) n=10;
  if(vgaxdim==800 && vgaydim==600) n=11;
  if(vgaxdim==1024 && vgaydim==768) n=12;
  if (n==0) { 
    fprintf(prot,"\nwarning: display: using default option -d640x480\n");
    fprintf(prot,"         you may select one out of -d320x240 -d640x480 -d800x600 -d1024x768,\n");
    fprintf(prot,"         depending on your svgalib setup and your hardware. \n"); 
    n=10; 
    vgaxdim=640; 
    vgaydim=480;
  }
  vgaydim-=STATUSLINE;

  if (!vga_hasmode(n)) 
    vgaerror("no such vga-mode on this machine");
  thegraphmode=n;
  pxlmem.bits= NULL;   /* important !!! */
}

void setpalette(void) {
  int i,gl;

  if(actgraphmode!=thegraphmode) return;
  gl_setcontextvga(actgraphmode);
  gl_setpalettecolor(BORDERCOL,0,40,10);  /* dark green */
  gl_setpalettecolor(TEXTCOL,0,63,16); /* light green */
  gl_setpalettecolor(TEXTBACKCOL,0,0,0); /* black */
  gl_setpalettecolor(FRAMECOL,40,10,5); /* some dark red */
  if(pxlmem.bits==NULL) return;
  if(pxlmem.type!=BLACKNWHITE) {
    for(i=0;i<COLORS_PER_GREY;i++) {
      gl=(i==0) ? 0: 63 * exp(GAMMA*log(i/(float) (COLORS_PER_GREY-1))) ;
      gl_setpalettecolor(COLORS_PER_GREY-1-i,gl,gl,gl);
    }
  } else {
    gl_setpalettecolor(0,63,63,63);
    gl_setpalettecolor(1,0,0,0);
  }
}

void vgaensgraph(void) {
  if(actgraphmode==thegraphmode) return;
  vga_setmode(thegraphmode);
  actgraphmode=thegraphmode;
  /* vgaxdim = vga_getxdim();
  vgaydim = vga_getydim()-STATUSLINE; */
  setpalette();
  preparefont();
}
 


void adjdvipos(void) {
  advixpos=dvixpos & (int)(-4);   /* make dvipos divisible by 4 */
  adviypos=dviypos;
}

#ifdef FLIPPIXELS
  static int cor[4]={0,1,-1,0};
#else
  static int cor[4]={0,0,0,0};
#endif

void copy2vga(int tx, int ty, int tw, int th, int sx, int sy) {
  
  int j,i,supi,supj,skip,ydest,xdest;
  register BMUNIT** bits=pxlmem.bits;
  BMUNIT mask, premask, firstpremask, data;
  long offset;

  /*fprintf(prot,"(copy2vga: tx %d ty %d tw %d th %d sx %d xy %d ",
                    tx,ty,tw,th,sx,sy); */
  if(pxlmem.type==BLACKNWHITE) {                               /* BlackNWhite*/
#ifdef USEMEMCPY
    skip= (sx & (BITS_PER_BMUNIT-1)) >> 2;  /* nibbels to skip */
    sx >>= BITS_LOG2;
    /*fprintf(prot,"tx %d ty %d tw %d th %d sx %d xy %d )\n", 
                    tx,ty,tw,th,sx,sy); */
    offset=ty*vgaxdim+tx;
    supi=sy+th;
    for (i = sy;i<supi;i++) {
#ifdef FLIPPIXELS
      mymemcpyflip1(offset,bits[i]+sx,tw,skip); 
#else 
      mymemcpystd1(offset,bits[i]+sx,tw,skip); 
#endif
      offset+=vgaxdim;
    }
#else
    skip= (sx & (BITS_PER_BMUNIT-1));  /* bits to skip */
    sx >>= BITS_LOG2;
    supi=sy+th;
    supj= sx + (tw  >> BITS_LOG2);
    premask = 1 << (BITS_PER_BMUNIT-1);
    firstpremask = premask >> skip;
    for (i = sy;i<supi;i++) {
      xdest=tx;
      mask = firstpremask;
      for (j = sx;j<supj;j++) {        
        data = bits[i][j];
        while(mask){
          gl_setpixel(xdest+cor[xdest & 3],ty,((data & mask) == 0) ? 0 : 1);
          mask >>= 1;
          xdest++;
	}
        mask = premask;
      }
      data = bits[i][j];
      while(xdest < tx+tw){
        gl_setpixel(xdest+cor[xdest & 3],ty,((data & mask) == 0) ? 0 : 1);
        mask >>= 1;
        xdest++;
      }
      ty++;
    }
#endif


  } else {                                                /* grey levels */
   

#ifdef USEMEMCPY
    supi=sy+th;
    offset=ty*vgaxdim+tx;
    sx>>= BITS_LOG2-BITS_PER_GREY_LOG2;
    for (i = sy;i<supi;i++) {
#ifdef FLIPPIXELS
      mymemcpyflip4(offset,bits[i]+sx,tw); 
#else
      mymemcpystd4(offset,bits[i]+sx,tw); 
#endif
      offset+= vgaxdim;
    }  
#else
    supi=sy+th;
    sx>>= BITS_LOG2-BITS_PER_GREY_LOG2;
    supj=sx+(tw >> (BITS_LOG2-BITS_PER_GREY_LOG2));
    for (i = sy;i<supi;i++) {
      xdest=tx;
      for (j = sx; j < supj; j++) {
        data=bits[i][j];
        gl_setpixel(xdest+cor[xdest & 3],ty,(data & 0xf000) >> 12); xdest++;
        gl_setpixel(xdest+cor[xdest & 3],ty,(data & 0xf00) >> 8); xdest++;
        gl_setpixel(xdest+cor[xdest & 3],ty,(data & 0xf0) >> 4); xdest++;
        gl_setpixel(xdest+cor[xdest & 3],ty,data & 0xf); xdest++;
      }
      ty++;
    }   
#endif
  }                                                         /* grey or not */
  /* fprintf(prot,")"); */
}

void vgatestdraw(int tx,int ty,int tw,int th) {
  
  int sx,sy,pxw;
  uchar lclip, rclip; 

  /*fprintf(prot,"(vgatestdraw: tx %d ty %d tw %d th %d ",tx,ty,tw,th); */
  adjdvipos();
  tx=MIN(MAX(0,tx),vgaxdim-1);
  ty=MIN(MAX(0,ty),vgaydim-1);
  tw=MIN(vgaxdim-tx,tw);
  th=MIN(vgaydim-ty,th);
  sx= advixpos+tx;
  sy= adviypos+ty;
  pxw= ((pxlmem.w -1) | 3) +1; /* round up to multiple of 4 */
  /*fprintf(prot,"B tx %d ty %d tw %d th %d sx %d sy %d ",tx,ty,tw,th,sx,sy);
  fflush(prot);*/
  rclip=0; lclip=0; 
  if((sx+tw<0) || (sx>pxw) || 
      (sy+th<0) || (sy>pxlmem.h) || (pxlmem.bits==NULL)) {
      gl_fillbox(tx,ty,tw,th,0);
      return;
  }
  if(sx<=0) {
    gl_fillbox(tx,ty,-sx,th,0);
    tx=tx-sx;
    tw=tw+sx;
    sx=0;
    lclip=1;
  }  
  if(sx+tw >= pxw) {
    gl_fillbox(tx+tw-sx-tw+pxw,ty,sx+tw-pxw,th,0);
    tw=pxw-sx;
    rclip=1;
  }
  if(sy<=0) {
    gl_fillbox(tx,ty,tw,-sy,0);
    ty=ty-sy;
    th=th+sy;
    sy=0;
    if((ty>0) && (tw>0))
      gl_hline(MAX(tx-1,0),ty-1,MIN(tx+tw,vgaxdim-1),BORDERCOL);
  }
  if(sy+th >= pxlmem.h) {
    gl_fillbox(tx,ty+th-sy-th+pxlmem.h,tw,sy+th-pxlmem.h,0);
    th=pxlmem.h-sy;
    if((ty+th<vgaydim) && (tw>0))
      gl_hline(MAX(tx-1,0),ty+th,MIN(tx+tw,vgaxdim-1),BORDERCOL);
  }
 
  /*fprintf(prot,"C tx %d ty %d tw %d th %d sx %d sy %d ",tx,ty,tw,th,sx,sy); 
  fflush(prot);*/
 
  if((tw>0) && (th>0)) copy2vga(tx,ty,tw,th,sx,sy);

  if(lclip && (tx>0)) 
    gl_line(tx-1,MAX(0,ty-1),tx-1,MIN(vgaydim-1,ty+th),BORDERCOL);
  if(rclip && (tx+tw<vgaxdim)) 
    gl_line(tx+tw,MAX(0,ty-1),tx+tw,MIN(vgaydim-1,ty+th),BORDERCOL);

  /* robos rand berechn- und malungen */
  if((tw>0) && (th>0) && frameon) {
    if((advixpos+tx <= lrandpxl) && (lrandpxl <= advixpos+tx+tw))
      gl_line(lrandpxl-advixpos,ty,lrandpxl-advixpos,ty+th-1,FRAMECOL);
    if((advixpos+tx <= pxw-rrandpxl-1) && (pxw-rrandpxl-1 <= advixpos+tx+tw))
      gl_line(pxw-rrandpxl-advixpos-1,ty,pxw-rrandpxl-advixpos-1,ty+th-1,
         FRAMECOL);
    if((adviypos+ty <= orandpxl) && (orandpxl <= adviypos+ty+th))
      gl_line(tx,orandpxl-adviypos,tx+tw-1,orandpxl-adviypos,FRAMECOL);
    if((adviypos+ty <= pxlmem.h-urandpxl-1) && 
         (pxlmem.h-urandpxl-1 <= adviypos+ty+th))
      gl_line(tx,pxlmem.h-urandpxl-adviypos-1,
        tx+tw-1,pxlmem.h-urandpxl-adviypos-1,FRAMECOL);
  }


 
  /*fprintf(prot,")"); */

} 

void vgaredraw(void) {
  char  s[MAXPATHSTR+50], ss[20];
  int i,j;

  vgaensgraph();
  vgatestdraw(0,0,vgaxdim,vgaydim);
  strcpy(s,"dvisvga: ");
  sprintf(ss," [%d]",cpagenum);
  i=vgaxdim / FONTSIZE - strlen(s)-strlen(ss)-strlen(dviname);
  j=strlen(s);
  for(;i>0;i--) s[j++] =' ';
  s[j]=0;
  i= -vgaxdim / FONTSIZE + strlen(s) + strlen(ss)+ strlen(dviname);
  strcat(s,dviname+i);
  strcat(s,ss);
  /*for(i=0;i<16;i++) s[i]=(char) 16*cpagenum+i;
  s[i]=0;  -- test the fonts */
  gl_fillbox(0,vgaydim,vgaxdim,STATUSLINE,TEXTBACKCOL);
  /* gl_write(FONTSIZE, vgaydim +STATUSLINE-FONTSIZE, s);  
          use this, when the bug in libvgagl is fixed */
  my_write(0, vgaydim +STATUSLINE-FONTSIZE, s);
}


void vgascroll(short dx,short dy) {
  int ox=advixpos;
  int oy=adviypos;
  

  if(abs(dx) > vgaxdim/4*3 || abs(dy) > vgaydim/4*3) {
    vgaredraw();
    return;
  }
  adjdvipos();
  dx=ox-advixpos;
  dy=oy-adviypos;
  if(dx<0) {
     gl_copybox(-dx,0,vgaxdim+dx,vgaydim,0,0);
     vgatestdraw(vgaxdim+dx,0,-dx,vgaydim);
  }
  if(dx>0) {
     gl_copybox(0,0,vgaxdim-dx,vgaydim, dx,0);
     vgatestdraw(0,0,dx,vgaydim);
  }   
  if(dy<0) {
     gl_copybox(0,-dy,vgaxdim,vgaydim+dy,0,0);
     vgatestdraw(0,vgaydim+dy,vgaxdim,-dy);
  }
  if(dy>0) {
     gl_copybox(0,0,vgaxdim, vgaydim-dy, 0,dy);
     vgatestdraw(0,0,vgaxdim,dy);
  }  
  /*vgaredraw();*/
}   





uchar vgacommand(void) {
  char input,next1,next2,next3;
  uchar com;

  input = vga_getkey();  
  next1=0; next2=0; next3=0;

  switch(input) { /* i know, this switching is a bit stupit ... */
   
    case 'q': case 'Q':  com= QUIT; break;
    case 'r': case 'R':  com= REDRAW; break;
    case 'c': case 'C':  com= LESSFINE; break;
    case 'f': case 'F':  com= MOREFINE; break;
    case 'g': case 'G':  com= GOTO; break;
    case 'o': case 'O':  com= TOGO; break;
    case 'p': case 'P':  com= TOFRAME; break;
    case 'z': case 'Z':  com= CENTER; break;
    case 'x': case 'X':  case 127: com= DELETE; break;
    case 0x2b: com= PLUS; break; 
    case 0x2d: com= MINUS; break;
 


    case 'm': case 'M': com= NEXT; break;
    case 'i': case 'I': com= PREV; break;
    case 'u': case 'U': com= UP; break;
    case 'n': case 'N': com= DOWN; break;
    case 'h': case 'H': com= LEFT; break;
    case 'j': case 'J': com= RIGHT; break;

    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9': 
      com=NUM0+(input-'0');
      break;
    case 27:
      next1=vga_getkey();
      switch(next1) {
      case 91:  
        next2=vga_getkey();
        switch(next2) {
        case 54: com= NEXT; break; 
        case 53: com= PREV; break;
        case 65: com= UP; break;
        case 66: com= DOWN; break;
        case 68: com= LEFT; break;
        case 67: com= RIGHT; break;
        case 71: com= CENTER; break;
        default: com= NOP;
	} break;
      default: com= NOP;
    } break;
    default: com= NOP;
  }
  /*if(input !=0 && input != -1)
    fprintf(prot,"(vgacommeand: %d %d %d %d)",input,next1,next2,next3);*/
  while(input !=0 && input != -1) input=vga_getkey();
  return(com);
}


/* font found in  vgagl ...
   i hate fonts. so i will use this one. it will be scaled, and pixels
   will be fliped. uaaa. and there is a tiny bug to fix, in
   vgagl/text.c, which skrews up compressed font writing. but i dont want
   to expand the font, cause using a 16x16 font this would take 64Kb of mem.

   this all meens: dont go on reading, unless youre the hearo, who
   writes all the mess new. may be using a pk-font ?
*/


/* binary image of font8x8 follows */
/* you may comment it out, cause its in vgalg */

static unsigned char __font8x8[] = {
    0,0,0,0,0,0,0,0,126,129,165,129,189,129,126,0, /* 0 to 15 */
    126,255,219,255,195,255,126,0,54,127,127,127,62,28,8,0, /* 16 to 31 */
    8,28,62,127,62,28,8,0,28,28,8,107,127,107,8,28, /* 32 to 47 */
    8,28,62,127,62,8,28,62,0,0,24,60,60,24,0,0, /* 48 to 63 */
    255,255,231,195,195,231,255,255,0,60,102,66,66,102,60,0, /* 64 to 79 */
    255,195,153,189,189,153,195,255,15,7,13,60,102,102,102,60, /* 80 to 95 */
    60,102,102,102,60,24,126,24,48,56,60,54,52,112,240,224, /* 96 to 111 */
    127,99,127,99,99,103,230,192,24,219,126,102,102,126,219,24, /* 112 to 127 */
    64,112,124,127,124,112,64,0,1,7,31,127,31,7,1,0, /* 128 to 143 */
    24,60,126,24,24,126,60,24,102,102,102,102,102,0,102,0, /* 144 to 159 */
    63,122,122,58,10,10,10,0,30,51,28,54,54,28,102,60, /* 160 to 175 */
    0,0,0,0,126,126,126,0,24,60,126,24,126,60,24,126, /* 176 to 191 */
    24,60,126,24,24,24,24,0,24,24,24,24,126,60,24,0, /* 192 to 207 */
    0,12,14,127,14,12,0,0,0,24,56,127,56,24,0,0, /* 208 to 223 */
    0,0,96,96,96,127,0,0,0,36,102,255,102,36,0,0, /* 224 to 239 */
    0,24,60,126,255,255,0,0,0,255,255,126,60,24,0,0, /* 240 to 255 */
    0,0,0,0,0,0,0,0,24,24,24,24,24,0,24,0, /* 256 to 271 */
    102,102,102,0,0,0,0,0,108,108,254,108,254,108,108,0, /* 272 to 287 */
    16,124,208,124,22,124,16,0,0,198,204,24,48,102,198,0, /* 288 to 303 */
    56,108,56,118,220,204,118,0,24,24,48,0,0,0,0,0, /* 304 to 319 */
    12,24,48,48,48,24,12,0,48,24,12,12,12,24,48,0, /* 320 to 335 */
    0,108,56,254,56,108,0,0,0,24,24,126,24,24,0,0, /* 336 to 351 */
    0,0,0,0,0,24,24,48,0,0,0,126,0,0,0,0, /* 352 to 367 */
    0,0,0,0,0,48,48,0,0,6,12,24,48,96,192,0, /* 368 to 383 */
    60,102,110,126,118,102,60,0,12,28,60,12,12,12,12,0, /* 384 to 399 */
    60,102,6,28,48,96,126,0,60,102,6,28,6,102,60,0, /* 400 to 415 */
    28,60,108,204,254,12,12,0,126,96,96,124,6,102,60,0, /* 416 to 431 */
    60,96,96,124,102,102,60,0,126,6,6,12,24,48,48,0, /* 432 to 447 */
    60,102,102,60,102,102,60,0,60,102,102,62,6,6,60,0, /* 448 to 463 */
    0,48,48,0,0,48,48,0,0,24,24,0,0,24,24,48, /* 464 to 479 */
    12,24,48,96,48,24,12,0,0,0,126,0,0,126,0,0, /* 480 to 495 */
    48,24,12,6,12,24,48,0,60,102,6,12,24,0,24,0, /* 496 to 511 */
    60,102,110,110,108,96,60,0,24,60,102,102,126,102,102,0, /* 512 to 527 */
    124,102,102,124,102,102,124,0,60,102,96,96,96,102,60,0, /* 528 to 543 */
    124,102,102,102,102,102,124,0,126,96,96,124,96,96,126,0, /* 544 to 559 */
    126,96,96,124,96,96,96,0,60,102,96,110,102,102,60,0, /* 560 to 575 */
    102,102,102,126,102,102,102,0,60,24,24,24,24,24,60,0, /* 576 to 591 */
    6,6,6,6,102,102,60,0,102,108,120,112,120,108,102,0, /* 592 to 607 */
    96,96,96,96,96,96,126,0,198,238,254,214,198,198,198,0, /* 608 to 623 */
    102,118,126,110,102,102,102,0,60,102,102,102,102,102,60,0, /* 624 to 639 */
    124,102,102,124,96,96,96,0,60,102,102,102,102,110,60,6, /* 640 to 655 */
    124,102,102,124,102,102,102,0,60,102,96,60,6,102,60,0, /* 656 to 671 */
    126,24,24,24,24,24,24,0,102,102,102,102,102,102,60,0, /* 672 to 687 */
    102,102,102,102,102,60,24,0,198,198,198,214,254,238,198,0, /* 688 to 703 */
    102,102,60,24,60,102,102,0,102,102,102,60,24,24,24,0, /* 704 to 719 */
    126,6,12,24,48,96,126,0,60,48,48,48,48,48,60,0, /* 720 to 735 */
    0,192,96,48,24,12,6,0,60,12,12,12,12,12,60,0, /* 736 to 751 */
    24,60,102,0,0,0,0,0,0,0,0,0,0,0,0,255, /* 752 to 767 */
    24,24,12,0,0,0,0,0,0,0,60,6,62,102,62,0, /* 768 to 783 */
    96,96,96,124,102,102,124,0,0,0,60,102,96,102,60,0, /* 784 to 799 */
    6,6,6,62,102,102,62,0,0,0,60,102,126,96,62,0, /* 800 to 815 */
    28,54,48,124,48,48,48,0,0,0,62,102,102,62,6,60, /* 816 to 831 */
    96,96,124,102,102,102,102,0,24,0,24,24,24,24,24,0, /* 832 to 847 */
    12,0,12,12,12,12,108,56,96,96,102,108,120,108,102,0, /* 848 to 863 */
    48,48,48,48,48,48,24,0,0,0,236,254,214,214,198,0, /* 864 to 879 */
    0,0,124,102,102,102,102,0,0,0,60,102,102,102,60,0, /* 880 to 895 */
    0,0,124,102,102,124,96,96,0,0,62,102,102,62,6,6, /* 896 to 911 */
    0,0,124,102,96,96,96,0,0,0,62,96,60,6,124,0, /* 912 to 927 */
    48,48,124,48,48,54,28,0,0,0,102,102,102,102,62,0, /* 928 to 943 */
    0,0,102,102,102,60,24,0,0,0,198,214,254,238,68,0, /* 944 to 959 */
    0,0,102,60,24,60,102,0,0,0,102,102,102,62,6,60, /* 960 to 975 */
    0,0,126,12,24,48,126,0,28,48,48,96,48,48,28,0, /* 976 to 991 */
    24,24,24,24,24,24,24,0,56,12,12,6,12,12,56,0, /* 992 to 1007 */
    118,220,0,0,0,0,0,0,0,0,24,60,102,102,126,0, /* 1008 to 1023 */
    60,102,96,96,102,60,24,48,102,0,102,102,102,102,62,0, /* 1024 to 1039 */
    14,0,60,102,126,96,60,0,60,102,60,6,62,102,62,0, /* 1040 to 1055 */
    102,0,60,6,62,102,62,0,112,0,60,6,62,102,62,0, /* 1056 to 1071 */
    24,24,60,6,62,102,62,0,0,0,62,96,96,62,24,48, /* 1072 to 1087 */
    60,102,60,102,126,96,60,0,102,0,60,102,126,96,60,0, /* 1088 to 1103 */
    112,0,60,102,126,96,60,0,102,0,24,24,24,24,24,0, /* 1104 to 1119 */
    60,102,24,24,24,24,24,0,112,0,24,24,24,24,24,0, /* 1120 to 1135 */
    198,56,108,198,254,198,198,0,24,24,0,60,102,126,102,0, /* 1136 to 1151 */
    14,0,124,96,120,96,124,0,0,0,126,26,126,216,126,0, /* 1152 to 1167 */
    62,120,216,222,248,216,222,0,60,102,60,102,102,102,60,0, /* 1168 to 1183 */
    102,0,60,102,102,102,60,0,112,0,60,102,102,102,60,0, /* 1184 to 1199 */
    60,102,0,102,102,102,62,0,112,0,102,102,102,102,62,0, /* 1200 to 1215 */
    102,0,102,102,102,62,6,60,102,60,102,102,102,102,60,0, /* 1216 to 1231 */
    102,0,102,102,102,102,60,0,12,12,62,96,96,62,12,12, /* 1232 to 1247 */
    56,108,96,240,96,102,252,0,102,102,60,24,126,24,126,24, /* 1248 to 1263 */
    124,102,102,124,102,111,102,99,14,27,24,60,24,24,120,48, /* 1264 to 1279 */
    14,0,60,6,62,102,62,0,14,0,24,24,24,24,24,0, /* 1280 to 1295 */
    14,0,60,102,102,102,60,0,14,0,102,102,102,102,62,0, /* 1296 to 1311 */
    118,220,0,124,102,102,102,0,126,0,102,118,126,110,102,0, /* 1312 to 1327 */
    62,102,102,62,0,126,0,0,60,102,102,60,0,126,0,0, /* 1328 to 1343 */
    24,0,24,48,96,102,60,0,0,0,0,126,96,96,0,0, /* 1344 to 1359 */
    0,0,0,126,6,6,0,0,198,204,216,62,99,198,12,31, /* 1360 to 1375 */
    198,204,216,54,110,214,31,6,24,0,24,24,24,24,24,0, /* 1376 to 1391 */
    0,54,108,216,108,54,0,0,0,216,108,54,108,216,0,0, /* 1392 to 1407 */
    34,136,34,136,34,136,34,136,85,170,85,170,85,170,85,170, /* 1408 to 1423 */
    221,119,221,119,221,119,221,119,8,8,8,8,8,8,8,8, /* 1424 to 1439 */
    8,8,8,8,248,8,8,8,8,8,8,248,248,8,8,8, /* 1440 to 1455 */
    28,28,28,28,252,28,28,28,0,0,0,0,252,28,28,28, /* 1456 to 1471 */
    0,0,0,248,248,8,8,8,28,28,28,252,252,28,28,28, /* 1472 to 1487 */
    28,28,28,28,28,28,28,28,0,0,0,252,252,28,28,28, /* 1488 to 1503 */
    28,28,28,252,252,0,0,0,28,28,28,28,252,0,0,0, /* 1504 to 1519 */
    8,8,8,248,248,0,0,0,0,0,0,0,248,8,8,8, /* 1520 to 1535 */
    8,8,8,8,15,0,0,0,8,8,8,8,255,0,0,0, /* 1536 to 1551 */
    0,0,0,0,255,8,8,8,8,8,8,8,15,8,8,8, /* 1552 to 1567 */
    0,0,0,0,255,0,0,0,8,8,8,8,255,8,8,8, /* 1568 to 1583 */
    8,8,8,15,15,8,8,8,28,28,28,28,31,28,28,28, /* 1584 to 1599 */
    28,28,28,31,31,0,0,0,0,0,0,31,31,28,28,28, /* 1600 to 1615 */
    28,28,28,255,255,0,0,0,0,0,0,255,255,28,28,28, /* 1616 to 1631 */
    28,28,28,31,31,28,28,28,0,0,0,255,255,0,0,0, /* 1632 to 1647 */
    28,28,28,255,255,28,28,28,8,8,8,255,255,0,0,0, /* 1648 to 1663 */
    28,28,28,28,255,0,0,0,0,0,0,255,255,8,8,8, /* 1664 to 1679 */
    0,0,0,0,255,28,28,28,28,28,28,28,31,0,0,0, /* 1680 to 1695 */
    8,8,8,15,15,0,0,0,0,0,0,15,15,8,8,8, /* 1696 to 1711 */
    0,0,0,0,31,28,28,28,28,28,28,28,255,28,28,28, /* 1712 to 1727 */
    8,8,8,255,255,8,8,8,8,8,8,8,248,0,0,0, /* 1728 to 1743 */
    0,0,0,0,15,8,8,8,255,255,255,255,255,255,255,255, /* 1744 to 1759 */
    0,0,0,0,255,255,255,255,240,240,240,240,240,240,240,240, /* 1760 to 1775 */
    15,15,15,15,15,15,15,15,255,255,255,255,0,0,0,0, /* 1776 to 1791 */
    0,0,118,204,204,204,118,0,60,102,102,124,102,102,124,96, /* 1792 to 1807 */
    126,102,96,96,96,96,96,0,0,0,254,108,108,108,102,0, /* 1808 to 1823 */
    126,102,48,24,48,102,126,0,0,0,62,108,108,108,56,0, /* 1824 to 1839 */
    0,0,102,102,102,102,127,192,0,0,126,216,24,24,12,0, /* 1840 to 1855 */
    124,56,124,214,214,124,56,124,124,198,198,254,198,198,124,0, /* 1856 to 1871 */
    124,198,198,198,108,108,238,0,30,48,24,60,102,102,60,0, /* 1872 to 1887 */
    0,0,126,219,219,126,0,0,3,6,62,107,115,62,96,192, /* 1888 to 1903 */
    30,48,96,126,96,48,30,0,124,198,198,198,198,198,198,0, /* 1904 to 1919 */
    0,126,0,126,0,126,0,0,24,24,126,24,24,0,126,0, /* 1920 to 1935 */
    48,24,12,24,48,0,126,0,12,24,48,24,12,0,126,0, /* 1936 to 1951 */
    14,27,27,24,24,24,24,24,24,24,24,24,24,216,216,112, /* 1952 to 1967 */
    24,24,0,126,0,24,24,0,0,118,220,0,118,220,0,0, /* 1968 to 1983 */
    60,102,102,60,0,0,0,0,0,0,0,24,24,0,0,0, /* 1984 to 1999 */
    0,0,0,0,24,0,0,0,30,24,24,24,24,216,120,56, /* 2000 to 2015 */
    120,108,108,108,108,0,0,0,56,12,24,48,60,0,0,0, /* 2016 to 2031 */
    0,0,60,60,60,60,0,0,0,0,0,0,0,0,0,0 /* 2032 to 2047 */
};

unsigned char *gl_font8x8 = __font8x8;

/* now the pronounced ugly fontscaling code: */

void preparefont(void) {
  unsigned char *dest;
  long i;
  short a,b,c;

#if (FONTSIZE == 16) 
  Tptr(font=(unsigned char *) malloc(2*16*256));
  dest=font;
  for(i=0;i<8*256;i++) {
    a= (short) gl_font8x8[i];
    b=0;
#ifndef FLIPPIXELS
    if((a & 0x01) !=0) b|=0x0003 ;
    if((a & 0x02) !=0) b|=0x000c ;
    if((a & 0x04) !=0) b|=0x0030 ;
    if((a & 0x08) !=0) b|=0x00c0 ;
    if((a & 0x10) !=0) b|=0x0300 ;
    if((a & 0x20) !=0) b|=0x0c00 ;
    if((a & 0x40) !=0) b|=0x3000 ;
    if((a & 0x80) !=0) b|=0xc000 ;
#else
    if((a & 0x01) !=0) b|=0x0005 ;
    if((a & 0x02) !=0) b|=0x000a ;
    if((a & 0x04) !=0) b|=0x0050 ;
    if((a & 0x08) !=0) b|=0x00a0 ;
    if((a & 0x10) !=0) b|=0x0500 ;
    if((a & 0x20) !=0) b|=0x0a00 ;
    if((a & 0x40) !=0) b|=0x5000 ;
    if((a & 0x80) !=0) b|=0xa000 ;
#endif
    *dest= (char) (b>>8);
    dest[2]=*dest;
    dest[1]=(char) (b & 0xff);
    dest[3]=dest[1];
    dest=dest+4;
  }
#else
#ifndef FLIPPIXELS
  font=gl_font8x8;
#else
  Tptr(font=(unsigned char *) malloc(8*256));
  dest=font;
  for(i=0;i<8*256;i++) {
    a=(short) gl_font8x8[i];
    c=(a & 0x99);
    if((a & 0x02) != 0) c|=0x04 ;
    if((a & 0x04) != 0) c|=0x02 ;
    if((a & 0x20) != 0) c|=0x40 ;
    if((a & 0x40) != 0) c|=0x20 ;
    *dest= (unsigned char) c;
    ++dest;
  }
#endif
#endif
  /* these 3 lines are to use gl_write if the bug is fixed some day */
  gl_setfontcolors(TEXTBACKCOL,TEXTCOL);
  gl_setwritemode(FONT_COMPRESSED);
  gl_setfont(FONTSIZE,FONTSIZE,font); 
}    


/* 
   from here on its a bugfix of a tiny bug in libvgagl. the bug results in
   libvgagl not being able to draw compressed fonts as they are. so when
   there is (soon ?) a fixed libvgagl, the following lines may be removed.
   i just did not want, to extract a whole font, just for drawing a few
   characters.
*/

static void my_expandcharacter( int bg, int fg, int c, unsigned char *bitmap ) {
	int x, y;
	unsigned char *exfont;
        /* here was the tiny bug: wrong braces in the next line */
	exfont = font + c * FONTSIZE * ((FONTSIZE + 7) / 8); 

	for (y = 0; y < FONTSIZE; y++)
		for (x = 0; x < FONTSIZE; x++) {
			int b;
			if (x % 8 == 0)
				b = *exfont++;
			if (b & (128 >> (x % 8)))	/* pixel */
				switch (BYTESPERPIXEL) {
				case 1 :
					*bitmap = fg;
					bitmap++;
				break;
				case 2 :
					*(ushort *)bitmap = fg;
					bitmap += 2;
					break;
				case 3 :
					*(ushort *)bitmap = fg;
					*(bitmap + 2) = fg >> 16;
					bitmap += 3;
					break;
				case 4 :
					*(uint *)bitmap = fg;
					bitmap += 4;
				}
			else	/* background pixel */
				switch (BYTESPERPIXEL) {
				case 1 :
					*bitmap = bg;
					bitmap++;
					break;
				case 2 :
					*(ushort *)bitmap = bg;
					bitmap += 2;
				break;
				case 3 :
					*(ushort *)bitmap = bg;
					*(bitmap + 2) = bg;
					bitmap += 3;
					break;
				case 4 :
					*(uint *)bitmap = bg;
					bitmap += 4;
				}
		}
}

/* Write using compressed font. */

static char bitmapbuffer[FONTSIZE * FONTSIZE];

void my_write( int x, int y, char *s ) {
	int i;
	i = 0;
	while (s[i] != 0) {
       		my_expandcharacter(TEXTBACKCOL,
	       		TEXTCOL, s[i], bitmapbuffer);
	       	gl_putbox(x + i * FONTSIZE, y, FONTSIZE,
		       	FONTSIZE, bitmapbuffer);
		i++;
	}
}


