/* This is part of dview, a dvi previewer. (c) 1995 Thomas Moor          */
/*                                                                       */
/* This program may be used without any warranty. It may be modified and */
/* distributed without any restrictions.                                 */

#define INCL_WIN
#define INCL_GPI

#include <os2.h>                        /* PM header file               */ 
#include <stdio.h>

#include "defsos2.h"
#include "writepm.h"                             

/* from the file globals.c we need ... */
extern FILE* prot;
extern char* dviname;
extern short cpagenum;
extern bigbitmap pxlmem;
extern int vgaxdim, vgaydim;
extern float lrandmm, rrandmm, orandmm, urandmm;
extern int lrandpxl, rrandpxl, orandpxl, urandpxl;
extern BMUNIT* bit_masks;
extern uchar* bit_count;
extern uchar* greytab;
extern uchar shrink;  /* stets <= 10 ? */
extern uchar colors;
extern int frameon;
extern int dvixpos, dviypos;

/* from the file subs.c w need ... */
extern void print_bigbitmap();
extern void* findmem(long);


typedef struct BINF {
  ULONG cbFix;
  ULONG cx,cy;
  USHORT cPlanes,cBitCount;
  ULONG ulCompression,cbImage,cxRes,cyRes,cclrUsed;
  RGB2 rgb[COLORS_PER_GREY];
} BINF;

/* globals belonging to the os2 and pm stuff */
HMQ  hmq = NULLHANDLE;                /* (only) Message queue handle     */
HWND hwndFrame = NULLHANDLE;          /* (only) Frame window handle      */
HWND hwndClient = NULLHANDLE;         /* (only) Client window handle      */
HAB  hab;                             /* PM anchor block handle       */ 

/* globals used by my painting stuff */
static char* vgapxlmem=NULL;
static BINF vgabinf;
static HBITMAP vgabmp=NULLHANDLE;
static HPS vgaps=NULLHANDLE;
static HDC vgadc=NULLHANDLE;
static int advixpos, adviypos;  

/* Function prototypes */
MRESULT EXPENTRY MyWindowProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 );


/* this became display dependent. the pxlmem should be accesable by the dvi-
   encoding part and by the routines that come with the display. */

void free_bigbitmap(void) {
  if(pxlmem.bits != NULL) free(pxlmem.bits);
  if(vgapxlmem != NULL) free(vgapxlmem);
  pxlmem.bits=NULL;
  vgapxlmem = NULL;
}

void alloc_bigbitmap(void) {
  short size, i;
  if(pxlmem.bits != NULL) free_bigbitmap();
  pxlmem.bits=(BMUNIT**) findmem(pxlmem.h*sizeof(BMUNIT*));
  memset(pxlmem.bits,0,pxlmem.h*sizeof(BMUNIT*));
  pxlmem.bmu_wide = ROUNDUP(pxlmem.w*pxlmem.type, BITS_PER_BMUNIT);
  size = MAX(pxlmem.bmu_wide*BYTES_PER_BMUNIT,1);
  size=ROUNDUP(size,sizeof(long int))*sizeof(long int);
  vgapxlmem=(char*)findmem(size*pxlmem.h);

  for(i=0;i<pxlmem.h;i++) 
    pxlmem.bits[pxlmem.h-i-1]= (BMUNIT*)(((char*)vgapxlmem)+i*size);

  vgabinf.cbFix=36L;
  vgabinf.cx=pxlmem.w;
  vgabinf.cy=pxlmem.h;
  vgabinf.cPlanes=1;
  vgabinf.cBitCount=(USHORT) pxlmem.type;
  vgabinf.ulCompression=BCA_UNCOMP;
  vgabinf.cbImage=size*pxlmem.h;
  vgabinf.cxRes=0;
  vgabinf.cyRes=0;
  vgabinf.cclrUsed=2;
  
  if(pxlmem.type!=BLACKNWHITE) {
    vgabinf.cclrUsed=COLORS_PER_GREY;
    for(i=0;i<COLORS_PER_GREY;i++) {
      size= 255 *(i/(float) (COLORS_PER_GREY-1))
              /*  *(i/(float) (COLORS_PER_GREY-1)) */ ;
      vgabinf.rgb[COLORS_PER_GREY-1-i].bRed=size;
      vgabinf.rgb[COLORS_PER_GREY-1-i].bGreen=size;
      vgabinf.rgb[COLORS_PER_GREY-1-i].bBlue=size;
      vgabinf.rgb[COLORS_PER_GREY-1-i].fcOptions=0;
    }
  } else {
    vgabinf.cclrUsed=2L;
    vgabinf.rgb[0].bRed=0;
    vgabinf.rgb[0].bGreen=0;
    vgabinf.rgb[0].bBlue=0;
    vgabinf.rgb[0].fcOptions=0;
    vgabinf.rgb[1].bRed=255;
    vgabinf.rgb[1].bGreen=255;
    vgabinf.rgb[1].bBlue=255;
    vgabinf.rgb[1].fcOptions=0;
  }
}

void vgaerror(char* mssg)
{
   PERRINFO  pErrInfoBlk;
   PSZ       pszOffSet;
   void      stdprint(void);

   fprintf(prot,"\nfatal error: display: %s\n");   
   if(hwndClient != NULLHANDLE) 
     WinPostMsg(hwndClient, WM_QUIT, (MPARAM)NULL, (MPARAM)NULL);
}



void vgaopen(void) {
  int n;
  uchar g;
  char title[MAXPATHSTR+20];
  ULONG flCreate;                       /* Window creation control flags*/
  POINTL pt1,pt2;
  RECTL rc;
  HPS hps;

  if ((hab = WinInitialize(0)) == 0L)   /* Initialize PM     */
     vgaerror("cant open pm");   /* Terminate the application    */

  if ((hmq = WinCreateMsgQueue( hab, 0 )) == 0L)/* Create a msg queue */
     vgaerror("cant open mspgq");   /* Terminate the application    */

  if (!WinRegisterClass(                /* Register window class        */
     hab,                               /* Anchor block handle          */
     (PSZ)"MyWindow",                   /* Window class name            */
     (PFNWP)MyWindowProc,               /* Address of window procedure  */
     CS_SIZEREDRAW,                     /* Class style                  */
     0                                  /* No extra window words        */
     ))
       vgaerror("cant open window"); /* Terminate the application    */

  flCreate = FCF_STANDARD &            /* Set frame control flags to   */
             ~FCF_SHELLPOSITION;        /* standard except for shell    */
                                        /* positioning.                 */

  if ((hwndFrame = WinCreateStdWindow(
               HWND_DESKTOP,            /* Desktop window is parent     */
               0,                       /* STD. window styles           */
               &flCreate,               /* Frame control flag           */
               "MyWindow",              /* Client window class name     */
               "",                      /* No window text               */
               0,                       /* No special class style       */
               (HMODULE)0L,           /* Resource is in .EXE file     */
               ID_WINDOW,               /* Frame window identifier      */
               &hwndClient              /* Client window handle         */
               )) == 0L)
       vgaerror("cant open window"); /* Terminate the application    */
  
  sprintf(title,"DVIew: %s",dviname);
  WinSetWindowText(hwndFrame, title);
    
    
  if (!WinSetWindowPos( hwndFrame,      /* Shows and activates frame    */
                   HWND_TOP,            /* window at default position  */
                   0, 0, vgaxdim, vgaydim,  /* and size.   */
                   SWP_SIZE |  SWP_ACTIVATE | SWP_SHOW
                 ))
    vgaerror("cant display window"); /* Terminate the application    */
  hps = WinBeginPaint( hwndClient, NULLHANDLE, &rc );
  pt1.x=0;
  pt2.x=vgaxdim-1;
  pt1.y=0;
  pt2.y=vgaydim-1;
  GpiSetColor(hps,CLR_DARKRED);
  GpiMove(hps,&pt1);
  GpiBox(hps,DRO_OUTLINEFILL,&pt2,0,0);
  WinEndPaint( hps );  
}

void vgaclose(void) {
  WinDestroyWindow(hwndFrame);           /* Tidy up...                   */
  WinDestroyMsgQueue(hmq);             /* Tidy up...                   */
  WinTerminate(hab);                   /* Terminate the application    */
}

void adjdvipos(void) {
  /*advixpos=dvixpos & 0xfff8; /* allign bytes. this should go fast ...  */
  advixpos=dvixpos | 1;        /* disallign bytes. this works ...  */
  adviypos=dviypos;
}

void vgascroll(short dx,short dy) {
  short ox=advixpos;
  short oy=adviypos;
  adjdvipos();
  dx=ox-advixpos;
  dy=adviypos-oy;
  WinScrollWindow(hwndClient,dx,dy,
        NULL,NULL,NULLHANDLE,NULL,SW_INVALIDATERGN); 
  /*WinScrollWindow(hwndClient,dx,-dy,NULL,NULL,NULLHANDLE,NULL,0); 
  WinInvalidateRect(hwndClient,NULL,FALSE);*/
}   


ushort cc=0;
void vgatestdraw(HPS hps, long t1x,long t1y,long t2x,long t2y) {
  POINTL abtl[4],pt1,pt2,pt3,pt4,pt5,pt6;
  long s1x,s1y,s2x,s2y;

  adjdvipos();
  s1x=advixpos+t1x;
  s1y=(long)vgabinf.cy-adviypos-vgaydim +  t1y;
  s2x=s1x+t2x-t1x+1;
  s2y=s1y+t2y-t1y+1;

  /*pt1.x=0;
  pt2.x=10;
  pt1.y=0;
  pt2.y=10;
  GpiSetColor(hps,CLR_RED);
  GpiMove(hps,&pt1);
  GpiBox(hps,DRO_OUTLINEFILL,&pt2,0,0);*/
  
  /*fprintf(prot,"\n(vgatestdraw: dx %d dy %d t1x %d t1y %d t2x %d t2y %d ",
         advixpos,adviypos,t1x,t1y,t2x,t2y);
  fprintf(prot,"\nsource: s1x %d s1y %d s2x %d s2y %d )",s1x,s1y,s2x,s2y);*/


  if((s2x+1<0L) || (s1x>(long) vgabinf.cx) || 
      (s2y+1 <0L) || (s1y > (long) vgabinf.cy) || (vgapxlmem==NULL)) {
    pt1.x=0;
    pt2.x=vgaxdim-1;
    pt1.y=0;
    pt2.y=vgaydim-1;
    GpiSetColor(hps,CLR_BACKGROUND);
    GpiMove(hps,&pt1);
    GpiBox(hps,DRO_OUTLINEFILL,&pt2,0,0);
    return;
  }
  pt1.x= -1;
  pt2.x= -1;
  pt3.x= -1;
  pt4.x= -1;
  if(s1x<=0) {
    pt1.x=t1x;
    pt2.x=t1x-s1x-2;
    pt1.y=t1y;
    pt2.y=t2y;
    t1x=pt2.x+2; 
    GpiSetColor(hps,CLR_BACKGROUND);
    GpiMove(hps,&pt1);
    GpiBox(hps,DRO_OUTLINEFILL,&pt2,0,0);
    pt1.x= ++ pt2.x;
  }  
  if(s2x>=(long) vgabinf.cx) {
    pt3.x=t2x-(s2x-(long)vgabinf.cx)+2;
    pt4.x=t2x;
    pt3.y=t1y;
    pt4.y=t2y;
    t2x=pt3.x-2;
    GpiSetColor(hps,CLR_BACKGROUND);
    GpiMove(hps,&pt3);
    GpiBox(hps,DRO_OUTLINEFILL,&pt4,0,0);
    pt4.x= --pt3.x;
  }
  if(s1y<=0) {
    pt5.x=t1x-1;
    pt6.x=t2x+1;
    pt5.y=t1y;
    pt6.y=t1y-s1y-2;
    t1y=pt6.y+2;
    GpiSetColor(hps,CLR_BACKGROUND);
    GpiMove(hps,&pt5);
    GpiBox(hps,DRO_OUTLINEFILL,&pt6,0,0);
    if(t2x-t1x >=0) {
      pt5.y= ++pt6.y; 
      GpiSetColor(hps,CLR_DARKGREEN);
      GpiMove(hps,&pt5);
      GpiLine(hps,&pt6);
    }
  }
  if(s2y>=(long)vgabinf.cy) {
    pt5.x=t1x-1;
    pt6.x=t2x+1;
    pt5.y=t2y-(s2y-(long)vgabinf.cy)+2;
    pt6.y=t2y;
    t2y=pt5.y-2;
    GpiSetColor(hps,CLR_BACKGROUND);
    GpiMove(hps,&pt5);
    GpiBox(hps,DRO_OUTLINEFILL,&pt6,0,0);
    if(t2x-t1x >= 0){
      pt6.y= --pt5.y;
      GpiSetColor(hps,CLR_DARKGREEN);
      GpiMove(hps,&pt5);
      GpiLine(hps,&pt6); 
    }
  }
  pt1.y=t1y-1;
  pt2.y=t2y+1;
  pt3.y=t1y-1;
  pt4.y=t2y+1;
  GpiSetColor(hps,CLR_DARKGREEN);
  GpiMove(hps,&pt1);
  GpiLine(hps,&pt2); 
  GpiMove(hps,&pt3);
  GpiLine(hps,&pt4); 

  s1x=advixpos+t1x;
  s1y=(long)vgabinf.cy-adviypos-vgaydim +  t1y;
  s2x=s1x+t2x-t1x+1;
  s2y=s1y+t2y-t1y+1;

  /*fprintf(prot,"\n(vgatestclip: dx %d dy %d t1x %d t1y %d t2x %d t2y %d ",
         advixpos,adviypos,t1x,t1y,t2x,t2y);
  fprintf(prot,"\nsource: s1x %d s1y %d s2x %d s2y %d )",s1x,s1y,s2x,s2y); */
 
  abtl[0].x=t1x;
  abtl[0].y=t1y;
  abtl[1].x=t2x;
  abtl[1].y=t2y;
  abtl[2].x=s1x;
  abtl[2].y=s1y;
  abtl[3].x=s2x;
  abtl[3].y=s2y;

  /*if(cc) GpiSetColor(hps,CLR_RED);
  else GpiSetColor(hps,CLR_GREEN);
  if(cc==0) GpiSetBackColor(hps,CLR_RED);
  else GpiSetBackColor(hps,CLR_GREEN);
  cc=1-cc;  
  GpiMove(hps,abtl);
  GpiBox(hps,DRO_OUTLINEFILL,abtl+1,0,0);
  GpiSetColor(hps,CLR_DEFAULT);
  GpiMove(hps,abtl);
  GpiBox(hps,DRO_OUTLINE,abtl+1,0,0); */ 
    
  if((t2x-t1x >=0) && (t2y-t1y >=0)){
    GpiSetColor(hps,CLR_DEFAULT);
    GpiDrawBits(hps,vgapxlmem,
           (PBITMAPINFO2)&vgabinf,4L,abtl,ROP_SRCCOPY,BBO_IGNORE); 
  }

  /*GpiWCBitBlt(hps,vgabmp,4L,abtl,ROP_SRCCOPY,BBO_IGNORE); */
  
  /*abtl[1].x++;
  abtl[1].y++;
  GpiBitBlt(hps,vgaps,3L,abtl,ROP_SRCCOPY,BBO_IGNORE); */
  
  /*pt1.x=0; 
  pt2.x=10;
  pt1.y=0;
  pt2.y=10;
  GpiSetColor(hps,CLR_GREEN);
  GpiMove(hps,&pt1);
  GpiBox(hps,DRO_OUTLINEFILL,&pt2,0,0);*/

  /* robos rand berechn- und malungen */ 
  if(frameon && (t2x-t1x >=0) && (t2y-t1y >=0)){
    GpiSetColor(hps,CLR_DARKRED);
    if((s1x <= lrandpxl) && (lrandpxl <= s2x)) { 
      pt1.x= t1x+lrandpxl-s1x;
      pt2.x= t1x+lrandpxl-s1x;
      pt1.y= t1y;
      pt2.y= t2y;
      GpiMove(hps,&pt1);
      GpiLine(hps,&pt2); 
    }
    if((s1x <= vgabinf.cx-1-rrandpxl) && (vgabinf.cx-1-rrandpxl <= s2x)) { 
      pt1.x= t1x+vgabinf.cx-1-rrandpxl-s1x;
      pt2.x= t1x+vgabinf.cx-1-rrandpxl-s1x;
      pt1.y= t1y;
      pt2.y= t2y;
      GpiMove(hps,&pt1);
      GpiLine(hps,&pt2); 
    }
    if((s1y <= urandpxl) && (urandpxl <= s2y)) { 
      pt1.x= t1x;
      pt2.x= t2x;
      pt1.y= t1y+urandpxl-s1y;
      pt2.y= t1y+urandpxl-s1y;
      GpiMove(hps,&pt1);
      GpiLine(hps,&pt2); 
    }
    if((s1y <= vgabinf.cy-1-orandpxl) && (vgabinf.cy-1-orandpxl <= s2y)) { 
      pt1.x= t1x;
      pt2.x= t2x;
      pt1.y= t1y+vgabinf.cy-1-orandpxl-s1y;
      pt2.y= t1y+vgabinf.cy-1-orandpxl-s1y;
      GpiMove(hps,&pt1);
      GpiLine(hps,&pt2); 
    }
  }

  /*fprintf(prot,"\n(last error: %x)",WinGetLastError(hab));*/
}      


void vgaredraw(void) {
   SIZEL siz;

   /*if(vgaps)  {
     GpiSetBitmap(vgaps,(HBITMAP) NULL);
     GpiDestroyPS(vgaps);
   }
   if(vgadc)  DevCloseDC(vgadc);
   if(vgabmp) GpiDeleteBitmap(vgabmp);
   if((vgadc=DevOpenDC(hab,OD_MEMORY,"*",0L,(PDEVOPENDATA) NULL,NULLHANDLE))
      ==NULLHANDLE) {
     vgaerror("bitmap creation error");
     return;
   }
   siz.cx=vgabinf.cx;
   siz.cy=vgabinf.cy;
   if((vgaps=GpiCreatePS(hab,vgadc,&siz,PU_PELS|GPIT_NORMAL|GPIA_ASSOC)
      )==NULLHANDLE) {
     vgaerror("bitmap creation error");
     return;
   }
   if((vgabmp= GpiCreateBitmap(vgaps,(PBITMAPINFOHEADER2)&vgabinf,
           CBM_INIT,vgapxlmem,(PBITMAPINFO2) &vgabinf)) == NULLHANDLE){
     vgaerror("bitmap creation error");
     return;
   }
   GpiSetBitmap(vgaps,vgabmp); */

   WinInvalidateRect(hwndClient,NULL,FALSE);
}



uchar vgacommand() {
  uchar com;  
  QMSG qmsg;                            /* Message from message queue   */
  WinGetMsg( hab, &qmsg, 0L, 0, 0 ); 
  com=(uchar) WinDispatchMsg( hab, &qmsg );
  return(com);
}


MRESULT EXPENTRY MyWindowProc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 ) {
  uchar res=FALSE;
  ushort input;
  switch( msg )
  {
    case WM_CREATE:
      break;     
    case WM_ERASEBACKGROUND:
      /*
       * Return TRUE to request PM to paint the window background
       * in SYSCLR_WINDOW.
       */
      res=FALSE;
      break;
    case WM_PAINT:
      /*
       * Window contents are drawn here in WM_PAINT processing.
       */
      {
      HPS    hps;                       /* Presentation Space handle    */
      RECTL   rc;
      char pagestring[MAXPATHSTR+13];
                                        /* Create a presentation space  */
      hps = WinBeginPaint( hwnd, NULLHANDLE, &rc );
      vgatestdraw(hps,rc.xLeft,rc.yBottom,rc.xRight,rc.yTop);
      sprintf(pagestring,"DVIew (%s [%d])",dviname,cpagenum);
      WinSetWindowText(hwndFrame, pagestring);  
      WinEndPaint( hps );                   
      break;
      }
    case WM_COMMAND:                    /* WM_COMMAND command value     */
      input = SHORT1FROMMP(mp1);      /* Extract the command value    */
      switch (input) {
      case ID_RIGHT:
        res=RIGHT; break;
      case ID_LEFT:
        res=LEFT; break;
      case ID_DOWN:
        res=DOWN; break;
      case ID_UP:
        res=UP; break;
      case ID_PAGEDOWN:
        res=NEXT; break;
      case ID_PAGEUP:
        res=PREV; break;
      case ID_BEGIN:
        res=FIRST; break;
      case ID_END:
        res=LAST; break;
      case ID_DELETE:
        res=DELETE; break;
      case ID_CTRLC:
        res=QUIT; break;
       }
      break;
   
    case WM_CHAR:
      if (!(SHORT1FROMMP(mp1) & KC_KEYUP)) 
      if (SHORT1FROMMP(mp1) & KC_CHAR){
        input=SHORT1FROMMP(mp2);
        switch(input) { 
          case 'q': case 'Q': res= QUIT; break;
          case 'r': case 'R': res= REDRAW; break;
          case '+': res= PLUS; break; 
          case '-': res= MINUS; break;
          case 'c': case 'C': res= LESSFINE; break;
          case 'f': case 'F': res= MOREFINE; break;
          case 'g': case 'G': res= GOTO; break;
          case 'o': case 'O': res= TOGO; break;
          case 'p': case 'P': res= TOFRAME; break;
          case 'z': case 'Z': res= CENTER; break;

	  case 'm': case 'M': res= NEXT; break;
	  case 'i': case 'I': res= PREV; break;
	  case 'u': case 'U': res= UP; break;
	  case 'n': case 'N': res= DOWN; break;
	  case 'h': case 'H': res= LEFT; break;
	  case 'j': case 'J': res= RIGHT; break;
          default: 
          if((0x30<=input) && (0x39 >= input)) res=NUM0+input-0x30;
          else res= NOP;
	}
      }
      break;
    case WM_SIZE:
         /* Window is being resized so get new client window size and   */
         /* recompute the end points for the lines.                     */
      vgaxdim = SHORT1FROMMP( mp2 );/* Get new client window size    */
      vgaydim = SHORT2FROMMP( mp2 );
      break;  
    case WM_CLOSE:
      /*
       * This is not the place to put my termination routines
       */
      WinPostMsg( hwnd, WM_QUIT, (MPARAM)0,(MPARAM)0 ); 
      res=QUIT;                                     /* Cause termination*/
      break;
    default:
      /*
       * Everything else comes here.  This call MUST exist
       * in your window procedure.
       */

      return(WinDefWindowProc( hwnd, msg, mp1, mp2 ));
  }
  return (MRESULT) res;
}




