/* This is part of dview, a dvi previewer. (c) 1995 Thomas Moor          */
/*                                                                       */
/* This program may be used without any warranty. It may be modified and */
/* distributed without any restrictions.                                 */


#include <stdio.h>
#include <math.h>
#include <string.h>
#include "defs.h"


/* from the file globals.c we need ... */

extern fontdesc **fontvect;
extern FILE* prot;
extern bigbitmap pxlmem;
extern BMUNIT  *bit_masks;
extern uchar *bit_count;
extern uchar shrink;
extern uchar colors;
extern char *dviname;
extern float  voffmm,hoffmm;
extern float papxmm,papymm;
extern float lrandmm, rrandmm, orandmm, urandmm;
extern int   lrandpxl, rrandpxl, orandpxl, urandpxl;
extern int xres,yres;
extern long newmag;
extern char *tfmprefix;
extern char *fontprefix;
extern int cpagenum;
extern uchar* greytab;



/* from the file subs.c we need ... */
extern long scaled(long ,long);
extern void* findmem(long int);
extern void Tptr(void*);
extern void fill_rect_big_bnw(short, short, short, short);
extern void fill_rect_big_gs(uchar, short,short,short,short);
extern void clear_bigbitmap(void);

/*from the file write*.c we need ... */
extern void alloc_bigbitmap(void);
extern void free_bigbitmap(void);

/* from the file readpk.c we need */
extern int pkfind(char*, char*, float, char*);
extern int pkdef(char*,int,long);
extern void pkloadchar(uchar,int);
extern void pkshrinkchar(chdesc*);

/* from the file readtfm.c we need */
extern int tfmfind(char*, char*, char*);
extern int tfmdef(char*,int,long);





FILE *dvifile=NULL;
char STR[MAXPATHSTR];

void dvibad(char* s) {
  fprintf(prot, "\nbad dvi: %s\n", s);  
  exit(1);
}

void dvibytes(char *s, uchar  l) {  
   s[l] = '\0';  
  if(fread(s,1,(long) l,dvifile)==0) dvibad("ends prematurely !");
}

uchar dvibyte() { 
  char rbuf[1];
  if(fread(rbuf, 1, 1, dvifile)==0) dvibad("ends prematurely !");
  return(rbuf[0]);
}

void dviskip(long l)  {  
  if(fseek(dvifile,l,1)!=0) dvibad("ends prematurely !");
} 

long dvinum(uchar b)  {  
  uchar i;  
  long n;
  uchar rbuf[4];
  if(fread(rbuf, b, 1, dvifile)==0) dvibad("ends prematurely !");
  n = 0;
  for (i = 1; i <= b; i++)  
    n = (n << 8) | (uchar) rbuf[i-1];  
  return n;
}

long dviint(uchar b)  {  
  uchar i;  
  long n;
  uchar rbuf[4];
  if(fread(rbuf, b, 1, dvifile)==0) dvibad("ends prematurely !");
 
  n = rbuf[0];
  if (n > 127)  n -= 256;
  for (i = 2; i <= b; i++)  
    n = (n << 8) | rbuf[i-1];  
  return n;
}

void dviposit(long l) {  
  if(fseek(dvifile, l, 0) !=0) dvibad("ends prematurely !") ;
}

long dvitell(void) {
  return(ftell(dvifile));
}

int dviopen(void) { 
  char s[MAXPATHSTR];
  int slen;
  
  strcpy(s,dviname);
  slen=strlen(s);
  if(slen==0) return(0);
  if(s[slen-1] == '.') strcat(s,"dvi");
  else if(strcasecmp(s+MAX(0,slen-4),".dvi") != 0) strcat(s,".dvi");
  dvifile = fopen(s, "rb");
  if(dvifile != NULL) 
    strcpy(dviname,s);
  else 
    dvifile= fopen(dviname, "rb");  
  return(dvifile != NULL);
}

void dviclose(void) {
  if(dvifile != NULL) 
    fclose(dvifile);
  dvifile=NULL;
}

void dvipositpost(void) {

  long l;  
  uchar j, em;  

  dviposit(0L);
  if (dvibyte() != 247) dvibad("no valid preamble");
  em = dvibyte();
  fseek(dvifile,0L,2);
  do {
    fseek(dvifile,-2L,1);
    j=dvibyte();
  } while (j == 223);
  if (j != em) dvibad("no valid postamble");
  fseek(dvifile, -5L, 1);
  l = dvinum(4);  
  fseek(dvifile, l, 0);  
  if (dvibyte() != 248) dvibad("no valid postamble");
}




/* local globals for dvi-file-processing */

static long hs[SMAX], vs[SMAX], ws[SMAX], xs[SMAX], ys[SMAX], zs[SMAX];
static long count[10];
static char comment[257], dvicomment[257];

static double conv, vconv, sconv, svconv;
static long  dvimag ,a, b, q, w, x, y, z;
static long  mag, spl,  mxh, mxv, lastpage;
static int pageanz, k, h, v, hoffpxl, voffpxl;
static ushort s, mst;

static long *pagelist=NULL;



void fontdef(long fontnr) { 
  uchar  p, q;
  int f,a;
  long dsz,scf,chksum;
  float fontmag;  
  fontdesc *font;
  char ffpath[MAXPATHSTR], fftuto[MAXPATHSTR];
  char fontname[MAXPATHSTR];

  if (fontnr < 0)  dvibad("negative fontnumber");
  f = fontnr & 0xff;  a = 0;
  if ((fontnr >= MAXFONTS) || (fontvect[f] != NULL)) do {
    if (fontvect[a] == NULL)  f = a;
    else if (fontvect[a]->dfn == fontnr)  dvibad("double fontnumber");
    a++;
  } while (a<MAXFONTS);
  if (fontvect[f] != NULL)  dvibad("too many fonts");
  fontvect[f] = (fontdesc *) findmem(sizeof(fontdesc));
  font = fontvect[f];  
  font->dfn = fontnr;
  chksum = dvinum(4);  
  scf = dvinum(4);  
  font->scf = scf;
  dsz = dvinum(4);  
  font->fty = NOTYPE;
  font->chv = NULL;
  font->fonam = NULL; 
  font->mch = -1;
  p = dvibyte(); 
  q = dvibyte();  
  dvibytes(fontname, q);
  if (p == 0)  strcpy(ffpath, fontprefix);  else dvibytes(ffpath, p);
  fontmag = xres * dvimag / 1000.0 * ((double)scf / dsz);
  if(pkfind(ffpath,fontname,fontmag,fftuto))
    if(pkdef(fftuto,f,chksum)) return;
  if(tfmfind(tfmprefix,fontname,fftuto))
    if(tfmdef(fftuto,f,chksum)) return;
  font->fonam=findmem(q+1);
  strcpy(font->fonam,fontname);
  
}


uchar getintnr(long fontnr) {   
  int a, f;
  f = fontnr & 0xff;  
  if (fontvect[f] != NULL) 
    if (fontvect[f]->dfn == fontnr)
      return(f);
  a = 0;
  do {
    if (fontvect[a] == NULL)	a++;
      else if (fontvect[a]->dfn != fontnr)  a++;
        else return(f);
  } while (a < MAXFONTS);
  fprintf(prot,"%d",fontnr);
  dvibad("is not defined as a fontnumber");
}


#define rnd(a) ((short)floor(a+0.5))
#define low(a) ((short)floor(a))

static void putrule() {  
  double le,bo,hi,wi;
  ushort gs;
  if ((b <= 0) || (a<=0))  return;
  le = sconv * h + hoffpxl;
  bo = svconv * v + voffpxl;
  hi = (svconv * a)-1;  
  wi = (sconv * b)-1;
  /*fprintf(prot,"(put_rule: le %5.1f bo %5.1f wi %5.1f hi %5.1f) ",le,bo,wi,hi);*/
  if((colors==BLACKNWHITE) || (shrink==1)) {
    wi=MAX(0,wi); hi=MAX(0,hi); /*enshure drawing at least something*/
    fill_rect_big_bnw(rnd(le), rnd(bo)-rnd(hi),rnd(le)+rnd(wi),rnd(bo));
    return;
  }
  fill_rect_big_gs(greytab[shrink*shrink],
           rnd(le), rnd(bo)-low(hi),rnd(le)+low(wi),rnd(bo));
  gs=(uchar)(shrink*shrink* (wi-low(wi))+0.49); /* 0.5 should bee ok ... */
  if(wi < 0) gs=MAX(1,gs);
  fill_rect_big_gs(greytab[gs],
        rnd(le)+low(wi)+1,rnd(bo)-low(hi),rnd(le)+low(wi)+1,rnd(bo));
  gs=(uchar)(shrink*shrink* (hi-low(hi))+0.49); /* but i hate seg.faults */
  if(hi < 0) gs=MAX(1,gs);
  fill_rect_big_gs(greytab[gs],
        rnd(le), rnd(bo)-low(hi)-1,rnd(le)+low(wi),rnd(bo)-low(hi)-1);
}

static void putchar_(uchar c, uchar fontintnr) {  
  fontdesc *font;  
  chdesc *thechar;
  short hr,vr,ar,br;

  font = fontvect[fontintnr];
  if((font->mch < c) || (font->fty == NOTYPE)) {
    fprintf(prot,
    "(putchar: warning: skipping undefined charakter %d in %s)",c,font->fonam);
    q = 0;
    return;
  } 
  thechar=&font->chv[c];
  if (font->fty == TFMTYPE) {   
    hr = (long)(sconv * h + 0.5) + hoffpxl;
    vr = (long)(svconv * (v - thechar->U2.tfh) + 0.5) + voffpxl;
    br = (long)(sconv * (h + thechar->U2.tfw) + 0.5) + hoffpxl;
    ar = (long)(svconv * (v + thechar->U2.tfd) + 0.5) + voffpxl;
    if((colors==BLACKNWHITE) || (shrink==1)) {
      fill_rect_big_bnw(hr, vr, br, vr);
      fill_rect_big_bnw(hr, vr, hr, ar);
      fill_rect_big_bnw(hr, ar, br, ar);
      fill_rect_big_bnw(br, vr, br, ar);
    } else {
      fill_rect_big_gs(greytab[shrink*shrink],hr, vr, br, vr);
      fill_rect_big_gs(greytab[shrink*shrink],hr, vr, hr, ar);
      fill_rect_big_gs(greytab[shrink*shrink],hr, ar, br, ar);
      fill_rect_big_gs(greytab[shrink*shrink],br, vr, br, ar);
    }
    q = thechar->U2.tfw;
    return;  
  } 
  /*fprintf(prot,"(putchar: pk: c %d fin %d h %d v %d",c,fontintnr,h,v);*/ 
  if(thechar->U1.bmp.bits == NULL) pkloadchar(c,fontintnr);
  if(shrink == 1) {
    hr = (long)(conv * h + 0.5) - thechar->U1.hof + hoffpxl;
    vr = (long)(vconv * v + 0.5) - thechar->U1.vof + voffpxl;
    copy_bigbitmap_bitmap(&thechar->U1.bmp,hr,vr);
  } else {
    if((thechar->U1.bmp2.bits == NULL) || (thechar->U1.shrink != shrink) 
        || (thechar->U1.bmp2.type != colors)) 
      pkshrinkchar(thechar);
    hr = (long)(sconv * h + 0.5) - thechar->U1.hof2 + hoffpxl;
    vr = (long)(svconv * v + 0.5) - thechar->U1.vof2 + voffpxl;
    copy_bigbitmap_bitmap(&thechar->U1.bmp2,hr,vr);
  } 
  q = thechar->U1.tfw;
  /*fprintf(prot,")");*/
}

static void pagebeg() {  
  uchar j;
  for (j = 0; j <= 9; j++)  count[j] = dvinum(4);
  dvinum(4);
  s = 0; 
  w = 0;  
  x = 0;  
  y = 0;  
  z = 0;    
  h = 0;
  v = 0;
}

static void pageend() {
  if (s != 0)  fprintf(prot, "warning: stack not empty at eop"); 
}

static void special() {  long i, j, k;  char *parm;
  double x, y, cy, ma, ye, bl;  short cle, m, n;
  fprintf(prot, "special >>");
  for (i = spl; i > 0; i--)  fputc(dvibyte(), prot);
  fprintf(prot, "<< ignored\n");

}

static void preamb(void) {  
  uchar l;
  long num,den;
  l = dvibyte();  
  num = dvinum(4);  
  den = dvinum(4);
  dvimag = dvinum(4);  
  if (newmag != 0) dvimag = newmag;
  conv = num / 254000.0 * ((double)xres / den) * (dvimag / 1000.0);
  vconv = num / 254000.0 * ((double)yres / den) * (dvimag / 1000.0);
  l = dvibyte();  dvibytes(dvicomment, l);  fprintf(prot, "%s\n", dvicomment);
}

static int testdvi(void) {
  uchar l;
  dviposit(14L); 
  l=dvibyte();
  dvibytes(comment, l);
  if(strcmp(comment,dvicomment)==0) return(1);
  else return(0);
}
  

static void postamb(void) {
  long num,den;
  lastpage=dvinum(4);  num = dvinum(4);  den = dvinum(4);
  dvimag = dvinum(4);  if (newmag != 0) dvimag = newmag;
  conv = num / 254000.0 * ((double)xres / den) * (dvimag / 1000.0);
  vconv = num / 254000.0 * ((double)yres / den) * (dvimag / 1000.0);
  mxv = dvinum(4);  mxh = dvinum(4);  mst = dvinum(2);  pageanz = dvinum(2);
}


static initpagelist(void) {
  short i;
  /*fprintf(prot,"(initpagelist: pagenaz %d",pageanz);*/
  pagelist=(long*) findmem(pageanz*sizeof(long));
  i=pageanz;
  pagelist[--i]=lastpage;
  while(i>0) {
    dviposit(pagelist[i]+ 1+4+(9*4));
    pagelist[--i]=dvinum(4);
  }
  /* fprintf(prot,")"); */
}    
  
  
short gotoprpage(short pn) {
  uchar o;
  pn=MAX(0,pn);
  pn=MIN(pageanz-1,pn);
  dviposit(pagelist[pn]);
  o= dvibyte();
  if(o!=139) dvibad("corruptet pagelist");
  pagebeg();
  return(pn);
}

void initdvi(void) { 
  int i,j,k;
  long f;
  uchar o;
 
  Tptr( fontvect = (fontdesc **) calloc(MAXFONTS, sizeof(fontdesc*)) );   
  
  bit_masks=(BMUNIT*) findmem(BYTES_PER_BMUNIT * (BITS_PER_BMUNIT+1));
  bit_masks[0]=0;
  for(j=1;j<=BITS_PER_BMUNIT;j++) 
      bit_masks[j] = (bit_masks[j-1] << 1) | 1;
  bit_count=(uchar*) findmem(256);
  for(j=0;j<256;) {
    k=0;
    for(i=0;i!=7;i++) k+= (j >> i) & 1; 
    bit_count[j++]=k;
  }


  if(!dviopen()) {
    fprintf(prot,"fatal error: cant open dvifile \n");
    exit(1);
  }

  dviposit(0L);
  if (dvibyte() != 247) dvibad("no valid preamble");
  preamb();

  dvipositpost();
  postamb();
  do {
    o = dvibyte(); 
    switch (o) {
      case 243:  case 244:  case 245:  case 246:
        f = dvinum(o - 242);  fontdef(f);  break;
      case 249: break;
      default: dvibad("there not only fontdefs in postamble ...");
    }
  } while(o != 249);

  initpagelist();
  dviclose();
}


void killdvi(void) {
  int i;
  uchar c;
  dviclose();
  if(fontvect!=NULL){
    for(i=0;i<MAXFONTS;i++) {
      if (fontvect[i] != NULL) {
        if (fontvect[i]->fty==PKTYPE) {
        for(c=0;c<=fontvect[i]->mch;c++) 
	  if(fontvect[i]->chv[c].U1.bmp.bits != NULL)
            free(fontvect[i]->chv[c].U1.bmp.bits);
          if(fontvect[i]->chv[c].U1.bmp2.bits != NULL) 
            free(fontvect[i]->chv[c].U1.bmp2.bits);
        }
        if(fontvect[i]->chv != NULL)
           free(fontvect[i]->chv);
        if(fontvect[i]->fonam != NULL) 
          free(fontvect[i]->fonam);
        free(fontvect[i]);
      }
    }
    free(fontvect);
    fontvect=NULL;
  }
  free_bigbitmap();
  if(greytab != NULL) {
    free(greytab);
    greytab=NULL;
  }
  if(bit_masks != NULL) {
   free(bit_masks);
   bit_masks = NULL;
  } 
  if(bit_count != NULL) {
   free(bit_count);
   bit_count = NULL;
  }
  if(pagelist != NULL) {
   free(pagelist);
   pagelist = NULL;
  }
}



int  dopage(short pn,uchar shr, uchar col){
  uchar o,c;
  int fint,i;
  ushort d1,d2;
  long f;
  if(!dviopen()) {
    fprintf(prot,"\nfatal error: dvifile disappeard\n");
    exit(1);
  }
  if(!testdvi()) {
    fprintf(prot,"\ndvifile has changed? reload\n");
    killdvi();
    initdvi();
    dviopen();
  }  
  shr=MAX(1,MIN(MAXSHRINK,shr));
  if((shr != shrink) || (col != colors) || (pxlmem.bits == NULL)) {
    shrink=shr;
    colors=col;
    if(shrink==1) pxlmem.type=BLACKNWHITE; else pxlmem.type=colors;
    free_bigbitmap();
    pxlmem.w=papxmm/25.4* xres /shrink;
    pxlmem.h=papymm/25.4* yres /shrink;
    alloc_bigbitmap();
    if(pxlmem.type == GREYSCALE) {
      if(greytab != NULL) free(greytab);
      greytab=(uchar*) findmem(shrink*shrink+1);
      for(i=0;i<=shrink*shrink;i++) 
        greytab[i]=ROUNDUP(i*(COLORS_PER_GREY-1),shrink*shrink);
    }
    hoffpxl=(long)(xres*hoffmm/25.4/shrink + 0.5);
    voffpxl=(long)(xres*voffmm/25.4/shrink + 0.5);
    lrandpxl=(long)(xres*lrandmm/25.4/shrink + 0.5);
    rrandpxl=(long)(xres*rrandmm/25.4/shrink + 0.5);
    orandpxl=(long)(xres*orandmm/25.4/shrink + 0.5);
    urandpxl=(long)(xres*urandmm/25.4/shrink + 0.5);
    pn=gotoprpage(pn);
  } else {
    pn=gotoprpage(pn);
    if(pn == cpagenum) {
      dviclose();
      return(0);
    }
  }
  cpagenum=pn;
  sconv=conv/shrink;
  svconv=vconv/shrink;
  fprintf(prot,"[%d",cpagenum);
  clear_bigbitmap();
  q=0;
  do {
    o = dvibyte();
    /*fprintf(prot,"(%d)",o); */
    switch (o) {
    case 128:  case 129:  case 130:  case 131:
      c = dvinum(o - 127);  putchar_(c,fint);  h += q;  break;
    case 132:  a = dviint(4);  b = dviint(4);
      putrule();  h += b;  break;
    case 133:  case 134:  case 135:  case 136:
      c = dvinum(o - 132);  putchar_(c,fint);  break;
    case 137:  a = dviint(4);  b = dviint(4);
      putrule();  break;
    case 138: break;
    case 139: dvibad("unexpectet bop"); break;
    case 140: break;
    case 141:  if (s < SMAX) {	s++;  hs[s - 1] = h;  vs[s - 1] = v;
	ws[s - 1] = w;	xs[s - 1] = x;	ys[s - 1] = y; zs[s - 1] = z;
      } else  dvibad("stack overflow");  break;
    case 142:  if (s > 0) {  h = hs[s - 1];  v = vs[s - 1];
	w = ws[s - 1];	x = xs[s - 1];	y = ys[s - 1];	z = zs[s - 1];	s--;
      } else  dvibad("stack underflow");  break;
    case 143:  case 144:  case 145:  case 146:
      h += dviint(o - 142);  break;
    case 147:  h += w;  break;
    case 148:  case 149:  case 150:  case 151:
      w = dviint(o - 147);  h += w;  break;
    case 152:  h += x;  break;
    case 153:  case 154:  case 155:  case 156:
      x = dviint(o - 152);  h += x;  break;
    case 157:  case 158:  case 159:  case 160:
      v += dviint(o - 156);  break;
    case 161:  v += y;  break;
    case 162:  case 163:  case 164:  case 165:
      y = dviint(o - 161);  v += y;  break;
    case 166:  v += z;  break;
    case 167:  case 168:  case 169:  case 170:
      z = dviint(o - 166);  v += z;  break;
    case 235:  case 236:  case 237:  case 238:
      f = dvinum(o - 234);  fint=getintnr(f);  break;
    case 239:  case 240:  case 241:  case 242:
      spl = dvinum(o - 238);  special();  break;
    case 243:  case 244:  case 245:  case 246:
      f = dvinum(o-242);
      /* fprintf(prot,"(fontdef:%d)",f); */
      dviskip(12L); 
      d1 = dvibyte();  d2 = dvibyte();  dviskip((long)(d1 + d2)); break;
    case 250:  case 251:  case 252:
    case 253:  case 254:  case 255:  
      fprintf(prot,"skipping unknown opc > 249 ");break;
    default:
      if (o <= 127) {
        putchar_(o,fint);  h += q;
      } else
        if (o >= 171 && o <= 234) {
	  fint=getintnr(o-171);
        } break;
    } /* end switch o */
  } while (o != 140);
  pageend();  
  /*print_bigbitmap();*/
  fprintf(prot,"] ");
  dviclose();
  return(1);
}
 







